'''
Created on 28.04.2016

@author: lemmerfn
'''
import numpy as np
import pysubgroup.utils as ut


class AbstractInterestingnessMeasure(object):
    def optimisticEstimateFromDataset(self, data, subgroup):
        return float("inf")

    def optimisticEstimateFromStatistics(self, instancesDataset, positivesDataset, instancesSubgroup,
                                         positivesSubgroup):
        return float("inf")

    def supportsWeights(self):
        return False

    def isApplicable(self, subgroup):
        return False


class BoundedInterestingnessMeasure:
    pass


class CombinedInterestingnessMeasure(AbstractInterestingnessMeasure, BoundedInterestingnessMeasure):
    def __init__(self, measures, weights=None):
        self.measures = measures
        if weights == None:
            weights = [1] * len(measures)
        self.weights = weights

    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):
        if not self.isApplicable(subgroup):
            raise BaseException("Quality measure cannot be used for this target class")
        return np.dot([m.evaluateFromDataset(data, subgroup, weightingAttribute) for m in self.measures], self.weights)

    def optimisticEstimateFromDataset(self, data, subgroup):
        if not self.isApplicable(subgroup):
            raise BaseException("Quality measure cannot be used for this target class")
        return np.dot([m.optimisticEstimateFromDataset(data, subgroup) for m in self.measures], self.weights)

    def evaluateFromStatistics(self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return np.dot(
            [m.evaluateFromStatistics(instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup) for m in
             self.measures], self.weights)

    def optimisticEstimateFromStatistics(self, instancesDataset, positivesDataset, instancesSubgroup,
                                         positivesSubgroup):
        return np.dot(
            [m.evaluateFromStatistics(instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup) for m in
             self.measures], self.weights)

    def isApplicable(self, subgroup):
        return all([x.isApplicable(subgroup) for x in self.measures])

    def supportsWeights(self):
        return all([x.supportsWeights() for x in self.measures])


##########
# Filter
##########

def uniqueAttributes(resultSet, data):
    result = []
    usedAttributes = []
    for (q, sg) in resultSet:
        atts = sg.subgroupDescription.getAttributes()
        if not atts in usedAttributes or all([ut.isCategoricalAttribute(data, x) for x in atts]):
            result.append((q, sg))
            usedAttributes.append(atts)
    return result


def minimumStatisticFilter(resultSet, statistic, minimum, data):
    result = []
    for (q, sg) in resultSet:
        if len(sg.statistics) == 0:
            sg.calculateStatistics(data)
        if sg.statistics[statistic] >= minimum:
            result.append((q, sg))
    return result


def minimumQualityFilter(resultSet, minimum):
    result = []
    for (q, sg) in resultSet:
        if q >= minimum:
            result.append((q, sg))
    return result


def maximumStatisticFilter(resultSet, statistic, maximum):
    result = []
    for (q, sg) in resultSet:
        if sg.statistics[statistic] <= maximum:
            result.append((q, sg))
    return result


def overlapFilter(resultSet, data, similarity_level=0.9):
    result = []
    resultSGs = []
    for (q, sg) in resultSet:
        if not overlapsList(sg, resultSGs, data, similarity_level):
            resultSGs.append(sg)
            result.append((q, sg))
    return result


def overlapsList(sg, listOfSGs, data, similarity_level=0.9):
    for anotherSG in listOfSGs:
        if ut.overlap(sg, anotherSG, data) > similarity_level:
            return True
    return False



'''
Created on 29.09.2017

@author: lemmerfn
'''
import numpy as np
import pysubgroup as ps
from functools import total_ordering
from pysubgroup.measures import AbstractInterestingnessMeasure, \
    BoundedInterestingnessMeasure
from pysubgroup.utils import conditional_invert



@total_ordering
class NumericTarget(object):
    def __init__(self, target_variable):
        self.target_variable = target_variable
        
    def __repr__(self):
        return "T: " + str(self.target_variable)
    
    def __eq__(self, other): 
        return self.__dict__ == other.__dict__
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
    def getAttributes(self):
        return [self.target_variable]
    
    def get_base_statistics (self, data, subgroup, weightingAttribute=None): 
        if (weightingAttribute is None):
            sg_instances = subgroup.subgroupDescription.covers(data)
            all_target_values = data[self.target_variable]
            sg_target_values = all_target_values[sg_instances]
            instances_dataset = len(data)
            instances_subgroup = np.sum(sg_instances)
            mean_sg = np.mean (sg_target_values)
            mean_dataset = np.mean (all_target_values)
            return (instances_dataset, mean_dataset, instances_subgroup, mean_sg)  
        else:
            raise NotImplemented("Attribute weights with numeric targets are not yet implemented.")
        
    def calculateStatistics (self, subgroup, data, weightingAttribute=None):
        if weightingAttribute is not None:
            raise NotImplemented("Attribute weights with numeric targets are not yet implemented.")
        sg_instances = subgroup.subgroupDescription.covers(data)
        all_target_values = data[self.target_variable]
        sg_target_values = all_target_values[sg_instances]
        subgroup.statistics['size_sg'] = len(sg_target_values)
        subgroup.statistics['size_dataset'] = len (data)
        subgroup.statistics['mean_sg'] = np.mean(sg_target_values)
        subgroup.statistics['mean_dataset'] = np.mean(all_target_values)
        subgroup.statistics['std_sg'] = np.std(sg_target_values)
        subgroup.statistics['std_dataset'] = np.std(all_target_values)
        subgroup.statistics['median_sg'] = np.median(sg_target_values)
        subgroup.statistics['median_dataset'] = np.median(all_target_values)
        subgroup.statistics['max_sg'] = np.max(sg_target_values)
        subgroup.statistics['max_dataset'] = np.max(all_target_values)
        subgroup.statistics['min_sg'] = np.min(sg_target_values)
        subgroup.statistics['min_dataset'] = np.min(all_target_values)
        subgroup.statistics['mean_lift'] = subgroup.statistics['mean_sg'] / subgroup.statistics['mean_dataset']
        subgroup.statistics['median_lift'] = subgroup.statistics['median_sg'] / subgroup.statistics['median_dataset']
        
class StandardQF_numeric (AbstractInterestingnessMeasure, BoundedInterestingnessMeasure):
    
    @staticmethod     
    def standardQF_numeric (a, instances_dataset, mean_dataset, instances_subgroup, mean_sg):
        if (instances_subgroup == 0):
            return 0
        return (instances_subgroup) ** a * (mean_sg - mean_dataset)
        
    def __init__(self, a, invert=False):
        self.a = a
        self.invert = invert
        
    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):
        if not self.isApplicable(subgroup):
            raise BaseException("Quality measure cannot be used for this target class")
        return ps.conditional_invert(self.evaluateFromStatistics (*subgroup.get_base_statistics(data, weightingAttribute)), self.invert)

    def optimisticEstimateFromDataset(self, data, subgroup):
        if not self.isApplicable(subgroup):
            raise BaseException("Quality measure cannot be used for this target class")
        sg_instances = subgroup.subgroupDescription.covers(data)
        all_target_values = data[subgroup.target.target_variable]
        mean_dataset = np.mean(all_target_values)
        sg_target_values = all_target_values[sg_instances]
        target_values_larger_than_mean = sg_target_values [sg_target_values > mean_dataset]
        return ps.conditional_invert(np.sum(target_values_larger_than_mean) - (len (target_values_larger_than_mean) * mean_dataset), self.invert)
        

    def evaluateFromStatistics (self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return StandardQF_numeric.standardQF_numeric (self.a, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup)
    
    def optimisticEstimateFromStatistics (self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return float("inf")

    def isApplicable(self, subgroup):
        return isinstance(subgroup.target, NumericTarget)

    def supportsWeights(self):
        return False

class GAStandardQF_numeric (AbstractInterestingnessMeasure):    
    def __init__(self, a, invert=False):
        self.a = a
        self.invert = invert
        
    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):
        (instances_dataset, _, instances_subgroup, mean_sg) = subgroup.get_base_statistics(data, weightingAttribute)
        if ((instances_subgroup == 0) or (instances_dataset == instances_subgroup)):
            return 0
        maxMean = getMaxGeneralizationMean(data, subgroup, weightingAttribute)
        relativeSize = (instances_subgroup / instances_dataset)
        return conditional_invert (relativeSize ** self.a * (mean_sg - maxMean), self.invert)

    def supportsWeights(self):
        return True

    def isApplicable(self, subgroup):
        return isinstance(subgroup.target, NumericTarget)
    
def getMaxGeneralizationMean(data, subgroup, weightingAttribute=None):
    selectors = subgroup.subgroupDescription.selectors
    generalizations = ps.powerset(selectors)
    maxMean = 0
    for sels in generalizations:
        sg = ps.Subgroup(subgroup.target, ps.SubgroupDescription(list(sels)))
        mean_sg = sg.get_base_statistics(data, weightingAttribute) [3]
        maxMean = max(maxMean, mean_sg)
    return maxMean


'''
Created on 19.10.2017

@author: lemmerfn
'''
def check_list(my_list,threshold):
    for i in my_list:
        if i <= threshold:
            my_list.remove(i)
    return my_list

print (check_list(list(range(1,12)), 5))
            


from .subgroup import *
from .algorithms import *
from .measures import *
from .utils import *
from .visualization import *

from .boolean_target import *
from .numeric_target import *
from .fi_target import *


''' 
Created on 28.04.2016

@author: lemmerfn
'''
import numpy as np
import pandas as pd
import pysubgroup.utils as ut
from functools import total_ordering


@total_ordering
class SubgroupDescription(object):
    def __init__(self, selectors):
        if isinstance(selectors, list) or isinstance(selectors, tuple):
            self.selectors = selectors
        else:
            self.selectors = [selectors]
    
    def covers(self, instance):
        # empty description ==> return a list of all '1's
        if (not self.selectors):
            return np.full(len(instance), True, dtype=bool)
        # non-empty description
        return np.all([sel.covers(instance) for sel in self.selectors], axis=0)

    def __len__(self):
        return len(self.selectors)

    def count (self, data):
        return sum(1 for x in data if self.covers(x))
    
    def getAttributes(self):
        return set([x.getAttributeName() for x in self.selectors])
   
    def to_string(self, open_brackets="", closing_brackets="", sel_open_bracket="", sel_closing_bracket="", and_term=" AND "):
        if not self.selectors:
            return "Dataset"
        result = open_brackets
        for sel in self.selectors:
            result += str(sel) + and_term
        result = result [:-len(and_term)]
        return result + closing_brackets
    
    def __repr__(self):
        return self.to_string()
   
    def __eq__(self, other): 
        return set(self.selectors) == set(other.selectors)
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
@total_ordering
class NominalSelector:
    def __init__(self, attributeName, attributeValue, name=None):
        self.attributeName = attributeName
        self.attributeValue = attributeValue
        self.selector_name = name

    def covers (self, data):
        return data[self.attributeName] == self.attributeValue
    
    def to_string(self, open_brackets="", closing_brackets=""):
        if self.selector_name == None:
            return open_brackets + str(self.attributeName) + "=" + str(self.attributeValue) + closing_brackets
        return open_brackets + self.selector_name + closing_brackets
    
    def __repr__(self):
        return self.to_string()
    
    def __eq__(self, other): 
        if None == other:
            return False
        return self.__dict__ == other.__dict__
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
    def __hash__(self): 
        return str(self).__hash__()
    
    def getAttributeName(self):
        return self.attributeName

@total_ordering
class NegatedSelector:
    def __init__(self, selector):
        self.selector = selector

    def covers (self, dataInstance):
        return not self.selector.covers(dataInstance)
    
    def __repr__(self):
        return self.to_string()
    
    def __eq__(self, other): 
        return self.__dict__ == other.__dict__
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
    def __hash__(self): 
        return str(self).__hash__()
    
    def to_string(self, open_brackets="", closing_brackets=""):
        return "NOT " + str(self.selector, open_brackets, closing_brackets)
    
    def getAttributeName(self):
        return self.selector.attributeName


# Including the lower bound, excluding the upperBound
@total_ordering
class NumericSelector:
    def __init__(self, attributeName, lowerBound, upperBound, name=None):
        self.attributeName = attributeName
        self.lowerBound = lowerBound
        self.upperBound = upperBound
        self.selector_name = name

    def covers (self, dataInstance):
        val = dataInstance[self.attributeName]
        return np.logical_and(val >= self.lowerBound, val < self.upperBound)
    
    def __repr__(self):
        return self.to_string()
    
    def __eq__(self, other): 
        if other == None:
            return False
        return self.__dict__ == other.__dict__
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
    def __hash__(self): 
        return str(self).__hash__()

    def to_string(self, open_brackets="", closing_brackets="", roundingDigits=2):
        formatter = "{0:." + str(roundingDigits) + "f}"
        ub = self.upperBound
        lb = self.lowerBound
        if ub % 1:
            ub = formatter.format(ub)
        if lb % 1:
            lb = formatter.format(lb)
        
        if self.selector_name != None:
            repre = self.selector_name
        elif self.lowerBound == float("-inf") and self.upperBound == float("inf"):
            repre = self.attributeName + "= anything"
        elif self.lowerBound == float("-inf"):
            repre = self.attributeName + "<" + str(ub)
        elif self.upperBound == float("inf"):
            repre = self.attributeName + ">=" + str(lb)
        else:
            repre = self.attributeName + ": [" + str(lb) + ":" + str(ub) + "["
        return open_brackets + repre + closing_brackets
    
    def getAttributeName(self):
        return self.attributeName 


@total_ordering
class Subgroup(object):
    def __init__(self, target, subgroupDescription):
        # If its already a NominalTarget object, we are fine, otherwise we create a new one
        # if (isinstance(target, NominalTarget) or isinstance(target, NumericTarget)):
        #    self.target = target
        # else:
        #    self.target = NominalTarget(target)
        
        # If its already a SubgroupDescription object, we are fine, otherwise we create a new one
        self.target = target
        if (isinstance(subgroupDescription, (SubgroupDescription))):
            self.subgroupDescription = subgroupDescription
        else:
            self.subgroupDescription = SubgroupDescription(subgroupDescription)
            
        # initialize empty cache for statistics
        self.statistics = {} 
    
    def __repr__(self):
        return "<<" + str(self.target) + "; D: " + str(self.subgroupDescription) + ">>"
    
    def covers(self, instance):
        return self.subgroupDescription.covers(instance)
    
    def count(self, data):
        return np.sum(self.subgroupDescription.covers(data))

    def __eq__(self, other):
        if other == None:
            return False
        return self.__dict__ == other.__dict__
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
    def get_base_statistics (self, data, weightingAttribute=None):
        return self.target.get_base_statistics (data, self, weightingAttribute)
    
    def calculateStatistics (self, data, weightingAttribute=None):
        self.target.calculateStatistics(self, data, weightingAttribute)
        

def createSelectors (data, nbins=5, intervals_only=True, ignore=None):
    if ignore is None:
        ignore = []
    sels = createNominalSelectors(data, ignore)
    sels.extend(createNumericSelectors(data, nbins, intervals_only, ignore=ignore))
    return sels

def createNominalSelectors(data, ignore=None):
    if ignore is None:
        ignore = []
    nominal_selectors = []
    for attr_name in [x for x in data.select_dtypes(exclude=['number']).columns.values if x not in ignore]:
        nominal_selectors.extend(createNominalSelectorsForAttribute(data, attr_name))
    return nominal_selectors

def createNominalSelectorsForAttribute(data, attribute_name):
    nominal_selectors = []
    for val in pd.unique(data[attribute_name]):
        nominal_selectors.append (NominalSelector(attribute_name, val))
    return nominal_selectors        
                    
def createNumericSelectors(data, nbins=5, intervals_only=True, weightingAttribute=None, ignore=None):
    if ignore is None:
        ignore = []
    numeric_selectors = []
    for attr_name in [x for x in data.select_dtypes(include=['number']).columns.values if x not in ignore]:
        numeric_selectors.extend(createNumericSelectorForAttribute(data, attr_name, nbins, intervals_only, weightingAttribute))
    return numeric_selectors

def createNumericSelectorForAttribute(data, attr_name, nbins=5, intervals_only=True, weightingAttribute=None):
            numeric_selectors = []
            uniqueValues = np.unique(data[attr_name])
            if (len(uniqueValues) <= nbins):
                for val in uniqueValues: 
                    numeric_selectors.append(NominalSelector(attr_name, val))
            else: 
                cutpoints = ut.equalFrequencyDiscretization(data, attr_name, nbins, weightingAttribute)
                if intervals_only:
                    old_cutpoint = float ("-inf")
                    for c in cutpoints:
                        numeric_selectors.append(NumericSelector(attr_name, old_cutpoint, c))
                        old_cutpoint = c
                    numeric_selectors.append(NumericSelector(attr_name, old_cutpoint, float("inf")))
                else:
                    for c in cutpoints:
                        numeric_selectors.append(NumericSelector(attr_name, c, float("inf")))
                        numeric_selectors.append(NumericSelector(attr_name, float("-inf"), c))
                
            return numeric_selectors

def removeTargetAttributes (selectors, target):
    result = []
    for sel in selectors:
        if not sel.getAttributeName() in target.getAttributes():
            result.append(sel)
    return result




from scipy.io import arff

import pysubgroup as ps
import pandas as pd
from timeit import default_timer as timer


data = pd.DataFrame (arff.loadarff("../data/credit-g.arff") [0])

target = ps.NominalTarget('class', b'bad')
searchSpace = ps.createNominalSelectors(data, ignore=['class'])
task = ps.SubgroupDiscoveryTask (data, target, searchSpace, resultSetSize=10, depth=5, qf=ps.ChiSquaredQF())


start = timer()
result = ps.BSD().execute(task)
end = timer()
print("Time elapsed: ", (end - start))
for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))


print ("******")
start = timer()
result = ps.TID_SD().execute(task, use_sets=True)
end = timer()
print("Time elapsed: ", (end - start))
for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))


print ("******")
start = timer()
result = ps.TID_SD().execute(task)
end = timer()
print("Time elapsed: ", (end - start))
for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))



from scipy.io import arff
import pysubgroup as ps
import pandas as pd


data = pd.DataFrame (arff.loadarff("../data/credit-g.arff") [0])

target = ps.NumericTarget ('credit_amount')
sg = ps.Subgroup (target, ps.NominalSelector("purpose", b"other"))
print (target.get_base_statistics(data, sg))
sg.calculateStatistics(data)
# pp.pprint (sg.statistics)

qf = ps.StandardQF_numeric (1.0)
print (qf.evaluateFromDataset(data, sg))



from scipy.io import arff

import pysubgroup as ps
import pandas as pd

data = pd.DataFrame (arff.loadarff("../data/credit-g.arff") [0])

target = ps.NominalTarget ('class', b'bad')
searchSpace = ps.createNominalSelectors(data, ignore=['class'])
task = ps.SubgroupDiscoveryTask (data, target, searchSpace, resultSetSize=10, depth=3, qf=ps.StandardQF(1.0))

result = ps.BeamSearch(beamWidth=10).execute(task)
for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))

print ("******")
result = ps.SimpleDFS().execute(task)
for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))

# print WRAccQF().evaluateFromDataset(data, Subgroup(target, []))



'''
Created on 29.04.2016

@author: lemmerfn
'''
import copy
from heapq import heappush, heappop
from itertools import islice
from typing import List

import pysubgroup.measures as m
import pysubgroup.utils as ut
from pysubgroup.subgroup import Subgroup, SubgroupDescription


class SubgroupDiscoveryTask(object):
    '''
    Capsulates all parameters required to perform standard subgroup discovery 
    '''
    def __init__(self, data, target, searchSpace, qf, resultSetSize=10, depth=3, minQuality=0, weightingAttribute=None):
            self.data = data
            self.target = target
            self.searchSpace = searchSpace
            self.qf = qf
            self.resultSetSize = resultSetSize
            self.depth = depth
            self.minQuality = minQuality
            self.weightingAttribute = weightingAttribute


class Apriori(object):
    def execute(self, task):
        measure_statistics_based = hasattr(task.qf, 'optimisticEstimateFromStatistics')
        result = []
        
        # init the first level
        next_level_candidates = []
        for sel in task.searchSpace:
            next_level_candidates.append (Subgroup(task.target, [sel]))
            
        # level-wise search
        depth = 1
        while (next_level_candidates):
            # check sgs from the last level
            promising_candidates = []
            for sg in next_level_candidates:
                if (measure_statistics_based):
                    statistics = sg.get_base_statistics(task.data)
                    ut.addIfRequired (result, sg, task.qf.evaluateFromStatistics (*statistics), task)
                    optimistic_estimate = task.qf.optimisticEstimateFromStatistics (*statistics) if isinstance(task.qf, m.BoundedInterestingnessMeasure) else float("inf")
                else:
                    ut.addIfRequired (result, sg, task.qf.evaluateFromDataset(task.data, sg), task)
                    optimistic_estimate = task.qf.optimisticEstimateFromDataset(task.data, sg) if isinstance(task.qf, m.BoundedInterestingnessMeasure) else float("inf")
                
                # optimistic_estimate = task.qf.optimisticEstimateFromDataset(task.data, sg) if isinstance(task.qf, m.BoundedInterestingnessMeasure) else float("inf") 
                # quality = task.qf.evaluateFromDataset(task.data, sg)
                # ut.addIfRequired (result, sg, quality, task)
                if (optimistic_estimate >= ut.minimumRequiredQuality(result, task)):
                    promising_candidates.append(sg.subgroupDescription.selectors)
            
            if (depth == task.depth):
                break
            
            # generate candidates next level
            next_level_candidates = []
            for i, sg1 in enumerate(promising_candidates):
                for j, sg2 in enumerate (promising_candidates):
                    if (i < j and sg1 [:-1] == sg2 [:-1]):
                        candidate = list(sg1) + [sg2[-1]]
                        # check if ALL generalizations are contained in promising_candidates
                        generalization_descriptions = [[x for x in candidate if x != sel] for sel in candidate]
                        if all (g in promising_candidates for g in generalization_descriptions):
                            next_level_candidates.append(Subgroup (task.target, candidate))
            depth = depth + 1
        
        result.sort(key=lambda x: x[0], reverse=True) 
        return result

class BestFirstSearch (object):
    def execute(self, task):
        result = []
        queue = []
        measure_statistics_based = hasattr(task.qf, 'optimisticEstimateFromStatistics')
        
        
        # init the first level
        for sel in task.searchSpace:
            queue.append ((float("-inf"), [sel]))
        
        while (queue):
            q, candidate_description = heappop(queue)
            q = -q
            if (q) < ut.minimumRequiredQuality(result, task):
                break
            
            sg = Subgroup (task.target, candidate_description)
            
            if (measure_statistics_based):
                statistics = sg.get_base_statistics(task.data)
                ut.addIfRequired (result, sg, task.qf.evaluateFromStatistics (*statistics), task)
                optimistic_estimate = task.qf.optimisticEstimateFromStatistics (*statistics) if isinstance(task.qf, m.BoundedInterestingnessMeasure) else float("inf")
            else:
                ut.addIfRequired (result, sg, task.qf.evaluateFromDataset(task.data, sg), task)
                optimistic_estimate = task.qf.optimisticEstimateFromDataset(task.data, sg) if isinstance(task.qf, m.BoundedInterestingnessMeasure) else float("inf")
            
            # compute refinements and fill the queue
            if (len (candidate_description) < task.depth and optimistic_estimate >= ut.minimumRequiredQuality(result, task)):
                # iterate over all selectors that are behind the last selector contained in the evaluated candidate according to the initial order
                index_of_last_selector = min (task.searchSpace.index(candidate_description[-1]), len (task.searchSpace) - 1)
                
                for sel in islice(task.searchSpace, index_of_last_selector + 1, None):
                    new_description = candidate_description + [sel]
                    heappush(queue, (-optimistic_estimate, new_description))
        result.sort(key=lambda x: x[0], reverse=True) 
        return result
    
class BeamSearch(object):
    '''
    Implements the BeamSearch algorithm. Its a basic implementation without any optimization, i.e., refinements get tested multiple times.
    '''
    def __init__(self, beamWidth=20, beamWidthAdaptive=False):
        self.beamWidth = beamWidth
        self.beamWidthAdaptive = beamWidthAdaptive
    
    def execute (self, task):
        # adapt beam width to the result set size if desired
        if self.beamWidthAdaptive:
            self.beamWidth = task.resultSetSize
        
        # check if beam size is to small for result set
        if (self.beamWidth < task.resultSetSize):
            raise RuntimeError('Beam width in the beam search algorithm is smaller than the result set size!')
        
        
        # init
        beam = [(0, Subgroup(task.target, []))]
        last_beam = None
        
        depth = 0
        while (beam != last_beam and depth < task.depth):
            last_beam = beam.copy()
            for (_, last_sg) in last_beam:
                for sel in task.searchSpace:
                    # create a clone
                    new_selectors = list(last_sg.subgroupDescription.selectors)
                    if not sel in new_selectors:
                        new_selectors.append(sel)
                        sg = Subgroup(task.target, new_selectors)
                        quality = task.qf.evaluateFromDataset (task.data, sg)
                        ut.addIfRequired(beam, sg, quality, task, check_for_duplicates=True)
            depth += 1

        result = beam [:task.resultSetSize]
        result.sort(key=lambda x: x[0], reverse=True) 
        return result


class SimpleDFS(object):
    def execute (self, task, useOptimisticEstimates=True):
        result = self.searchInternal(task, [], task.searchSpace, [], useOptimisticEstimates)
        result.sort(key=lambda x: x[0], reverse=True)
        return result


    def searchInternal(self, task: SubgroupDiscoveryTask, prefix: List, modificationSet: List, result: List, useOptimisticEstimates: bool) -> List:
        sg = Subgroup(task.target, SubgroupDescription(copy.copy(prefix)))
        optimisticEstimate = float("inf")
        if useOptimisticEstimates and len(prefix) < task.depth and isinstance(task.qf, m.BoundedInterestingnessMeasure):
            optimisticEstimate = task.qf.optimisticEstimateFromDataset(task.data, sg)
            if (optimisticEstimate <= ut.minimumRequiredQuality(result, task)):
                return result
        
        if task.qf.supportsWeights():
            quality = task.qf.evaluateFromDataset(task.data, sg, task.weightingAttribute)
        else: 
            quality = task.qf.evaluateFromDataset(task.data, sg)
        ut.addIfRequired (result, sg, quality, task)
     
        if (len(prefix) < task.depth):
            newModificationSet = copy.copy(modificationSet)
            for sel in modificationSet:
                prefix.append(sel)
                newModificationSet.pop(0)
                self.searchInternal(task, prefix, newModificationSet, result, useOptimisticEstimates)
                # remove the sel again
                prefix.pop(-1)
        return result

class BSD (object):
    """
    Implementation of the BSD algorithm for binary targets. See
    Lemmerich, Florian, Mathias Rohlfs, and Martin Atzmueller. "Fast Discovery of Relevant Subgroup Patterns." FLAIRS Conference. 2010.
    """
    def execute(self, task):
        self.popSize = len(task.data)

        # generate target bitset
        self.targetBitset = 0
        x = task.target.covers(task.data)
        for index, e in enumerate(x):
            self.targetBitset += int(e) << index
        self.popPositives = ut.count_bits(self.targetBitset)

        # generate selector bitsets
        self.bitsets = {}
        for sel in task.searchSpace:
            # generate bitset
            x = sel.covers(task.data).values
            selBitset = 0
            for index, e in enumerate(x):
                selBitset += int(e) << index
            self.bitsets[sel] = selBitset
        result = self.searchInternal(task, [], task.searchSpace, [], (1 << self.popSize) - 1)
        result.sort(key=lambda x: x[0], reverse=True)
        return result

    def searchInternal(self, task, prefix, modificationSet, result, bitset):

        sgSize = ut.count_bits(bitset)
        positiveInstances = bitset & self.targetBitset
        sgPositiveCount = ut.count_bits(positiveInstances)

        optimisticEstimate = task.qf.optimisticEstimateFromStatistics(self.popSize, self.popPositives, sgSize,
                                                                      sgPositiveCount)
        if (optimisticEstimate <= ut.minimumRequiredQuality(result, task)):
            return result

        sg = Subgroup(task.target, copy.copy(prefix))

        quality = task.qf.evaluateFromStatistics(self.popSize, self.popPositives, sgSize, sgPositiveCount)
        ut.addIfRequired(result, sg, quality, task)

        if (len(prefix) < task.depth):
            newModificationSet = copy.copy(modificationSet)
            for sel in modificationSet:
                prefix.append(sel)
                newBitset = bitset & self.bitsets[sel]
                newModificationSet.pop(0)
                self.searchInternal(task, prefix, newModificationSet, result, newBitset)
                # remove the sel again
                prefix.pop(-1)
        return result

class TID_SD (object):
    """
    Implementation of a depth-first-search with look-ahead using vertical ID lists as data structure.
    """

    def execute(self, task, use_sets=False):
        self.popSize = len(task.data)

        # generate target bitset
        if use_sets:
            self.targetBitset = set()
            for index, row in task.data.iterrows():
                if task.target.covers(row):
                    self.targetBitset.add(index)
        else:
            self.targetBitset = []
            for index, row in task.data.iterrows():
                if task.target.covers(row):
                    self.targetBitset.append(index)


        self.popPositives = len(self.targetBitset)

        # generate selector bitsets
        self.bitsets = {}
        for sel in task.searchSpace:
            # generate bitset
            if use_sets:
                selBitset = set()
                for index, row in task.data.iterrows():
                    if sel.covers(row):
                        selBitset.add(index)
            else:
                selBitset = []
                for index, row in task.data.iterrows():
                    if sel.covers(row):
                        selBitset.append(index)
            self.bitsets[sel] = selBitset
        if use_sets:
            result = self.searchInternal(task, [], task.searchSpace, [], set(range(self.popSize)), use_sets)
        else:
            result = self.searchInternal(task, [], task.searchSpace, [], list(range(self.popSize)), use_sets)
        result.sort(key=lambda x: x[0], reverse=True)
        return result

    def searchInternal(self, task, prefix, modificationSet, result, bitset, use_sets):

        sgSize = len(bitset)
        if use_sets:
            positiveInstances = bitset & self.targetBitset
        else:
            positiveInstances = ut.intersect_of_ordered_list(bitset, self.targetBitset)
        sgPositiveCount = len(positiveInstances)

        optimisticEstimate = task.qf.optimisticEstimateFromStatistics(self.popSize, self.popPositives, sgSize,
                                                                      sgPositiveCount)
        if (optimisticEstimate <= ut.minimumRequiredQuality(result, task)):
            return result

        sg = Subgroup(task.target, copy.copy(prefix))

        quality = task.qf.evaluateFromStatistics(self.popSize, self.popPositives, sgSize, sgPositiveCount)
        ut.addIfRequired(result, sg, quality, task)

        if (len(prefix) < task.depth):
            newModificationSet = copy.copy(modificationSet)
            for sel in modificationSet:
                prefix.append(sel)
                if use_sets:
                    newBitset = bitset & self.bitsets[sel]
                else:
                    newBitset = ut.intersect_of_ordered_list(bitset, self.bitsets[sel])
                newModificationSet.pop(0)
                self.searchInternal(task, prefix, newModificationSet, result, newBitset, use_sets)
                # remove the sel again
                prefix.pop(-1)
        return result




from scipy.io import arff

import pysubgroup as ps
import pandas as pd
from timeit import default_timer as timer

data = pd.DataFrame (arff.loadarff("../data/credit-g.arff") [0])

target = ps.NominalTarget ('class', b'bad')
searchSpace = ps.createSelectors(data, ignore=['class'])
task = ps.SubgroupDiscoveryTask (data, target, searchSpace, resultSetSize=10, depth=3, qf=ps.ChiSquaredQF(direction="bidirect"))

start = timer()
result = ps.SimpleDFS().execute(task)
end = timer()

print("Time elapsed: ", (end - start)) 

for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))   

# print WRAccQF().evaluateFromDataset(data, Subgroup(target, []))



'''
Created on 10.05.2017

@author: lemmerfn
'''
import pandas as pd
import pysubgroup as ps

if __name__ == '__main__':
    data = pd.read_csv("~/datasets/titanic.csv")
    target = ps.NominalSelector ('survived', 0)
    
    s1 = ps.Subgroup (target, [])
    s2 = ps.Subgroup (target, [])
    
    print (s1 == s2)


import pysubgroup as ps
import pandas as pd


data = pd.read_csv("~/datasets/titanic.csv")
target = ps.NominalTarget ('survived', 0)
searchSpace = ps.createSelectors(data, ignore=['survived'])
task = ps.SubgroupDiscoveryTask (data, target, searchSpace,
                                 resultSetSize=5, depth=2,
                                 qf=ps.CombinedInterestingnessMeasure([ps.StandardQF(1), ps.GAStandardQF(1)]))

result = ps.SimpleDFS().execute(task, useOptimisticEstimates=False)

for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))



'''
Created on 29.09.2017

@author: lemmerfn
'''
import numpy as np
import scipy.stats
from functools import total_ordering
from pysubgroup.measures import AbstractInterestingnessMeasure, \
    BoundedInterestingnessMeasure
import pysubgroup.utils as ut
from pysubgroup.subgroup import SubgroupDescription, Subgroup, NominalSelector

@total_ordering
class NominalTarget(object):
    
    def __init__(self, target_attribute=None, target_value=None, target_selector=None):
        """
        Creates a new target for the boolean model class (classic subgroup discovery). 
        If target_attribute and target_value are given, the target_selector is computed using attribute and value
        """
        if target_attribute is not None and target_value is not None:
            if target_selector is not None:
                raise BaseException("NominalTarget is to be constructed EITHER by a selector OR by attribute/value pair")
            target_selector = NominalSelector(target_attribute, target_value)
        if target_selector is None:
            raise BaseException("No target selector given")
        self.targetSelector = target_selector
        
    def __repr__(self):
        return "T: " + str(self.targetSelector)
    
    def __eq__(self, other): 
        return self.__dict__ == other.__dict__
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
    def covers(self, instance):
        return self.targetSelector.covers(instance)
    
    def getAttributes(self):
        return [self.targetSelector.getAttributeName()]

    def get_base_statistics (self, data, subgroup, weightingAttribute=None): 
        if weightingAttribute is None:
            sgInstances = subgroup.subgroupDescription.covers(data)
            positives = subgroup.target.covers(data)
            instancesSubgroup = np.sum(sgInstances)
            positivesDataset = np.sum(positives)
            instancesDataset = len(data)
            positivesSubgroup = np.sum(np.logical_and(sgInstances, positives))
            return (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup)  
        else:
            weights = data[weightingAttribute]
            sgInstances = subgroup.subgroupDescription.covers(data)
            positives = subgroup.target.covers(data)                         
    
            instancesDataset = np.sum(weights)
            instancesSubgroup = np.sum(np.dot(sgInstances, weights))
            positivesDataset = np.sum(np.dot(positives, weights))
            positivesSubgroup = np.sum(np.dot(np.logical_and(sgInstances, positives), weights))
            return (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup)
    
    def calculateStatistics (self, subgroup, data, weightingAttribute=None):
        (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup) = self.get_base_statistics (data, subgroup, weightingAttribute)
        subgroup.statistics['size_sg'] = instancesSubgroup
        subgroup.statistics['size_dataset'] = instancesDataset
        subgroup.statistics['positives_sg'] = positivesSubgroup
        subgroup.statistics['positives_dataset'] = positivesDataset
        
        subgroup.statistics['size_complement'] = instancesDataset - instancesSubgroup
        subgroup.statistics['relative_size_sg'] = instancesSubgroup / instancesDataset
        subgroup.statistics['relative_size_complement'] = (instancesDataset - instancesSubgroup) / instancesDataset
        subgroup.statistics['coverage_sg'] = positivesSubgroup / positivesDataset
        subgroup.statistics['coverage_complement'] = (positivesDataset - positivesSubgroup) / positivesDataset
        subgroup.statistics['target_share_sg'] = positivesSubgroup / instancesSubgroup
        subgroup.statistics['target_share_complement'] = (positivesDataset - positivesSubgroup) / (instancesDataset - instancesSubgroup)
        subgroup.statistics['target_share_dataset'] = positivesDataset / instancesDataset
        subgroup.statistics['lift'] = (positivesSubgroup / instancesSubgroup) / (positivesDataset / instancesDataset)
        
        if (weightingAttribute is not None):
            (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup) = subgroup.get_base_statistics(data, self, weightingAttribute)
        subgroup.statistics['size_sg_weighted'] = instancesSubgroup
        subgroup.statistics['size_dataset_weighted'] = instancesDataset
        subgroup.statistics['positives_sg_weighted'] = positivesSubgroup
        subgroup.statistics['positives_dataset_weighted'] = positivesDataset
        
        subgroup.statistics['size_complement_weighted'] = instancesDataset - instancesSubgroup
        subgroup.statistics['relative_size_sg_weighted'] = instancesSubgroup / instancesDataset
        subgroup.statistics['relative_size_complement_weighted'] = (instancesDataset - instancesSubgroup) / instancesDataset
        subgroup.statistics['coverage_sg_weighted'] = positivesSubgroup / positivesDataset
        subgroup.statistics['coverage_complement_weighted'] = (positivesDataset - positivesSubgroup) / positivesDataset
        subgroup.statistics['target_share_sg_weighted'] = positivesSubgroup / instancesSubgroup
        subgroup.statistics['target_share_complement_weighted'] = (positivesDataset - positivesSubgroup) / (instancesDataset - instancesSubgroup)
        subgroup.statistics['target_share_dataset_weighted'] = positivesDataset / instancesDataset
        subgroup.statistics['lift_weighted'] = (positivesSubgroup / instancesSubgroup) / (positivesDataset / instancesDataset)
        
    
class ChiSquaredQF (AbstractInterestingnessMeasure):
    @staticmethod     
    def chiSquaredQF (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup, min_instances=5, bidirect=True, direction_positive=True):
        if (instancesSubgroup < min_instances) or ((instancesDataset - instancesSubgroup) < min_instances):
            return float("-inf")
        p_subgroup = positivesSubgroup / instancesSubgroup
        p_dataset = positivesDataset / instancesDataset
        positivesComplement = positivesDataset - positivesSubgroup

        # instancesComplement = instancesDataset - instancesSubgroup
        negativesSubgroup = instancesSubgroup - positivesSubgroup
        negativesDataset = instancesDataset - positivesDataset
        negativesComplement = negativesDataset - negativesSubgroup
        
        # observed = [positivesSubgroup, positivesComplement,negativesSubgroup, negativesComplement]
        #
        # if round(positivesSubgroup) < 0 or round(positivesComplement) < 0 or round(negativesSubgroup) <0 or round (negativesComplement) < 0:
        #    print ("XXXXX")
        val = scipy.stats.chi2_contingency([[round(positivesSubgroup), round(positivesComplement)],
                                            [round(negativesSubgroup), round(negativesComplement)]], correction=False)[0]
        if bidirect:
            return val
        elif direction_positive and p_subgroup > p_dataset:
            return val
        elif not direction_positive and p_subgroup < p_dataset:
            return val
        return -val
    
    @staticmethod     
    def chiSquaredQFWeighted (subgroup, data, weightingAttribute, effectiveSampleSize=0, min_instances=5,):
        (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup) = subgroup.get_base_statistics(data, weightingAttribute)
        if (instancesSubgroup < min_instances) or ((instancesDataset - instancesSubgroup) < 5):
            return float("inf")
        if effectiveSampleSize == 0:
            effectiveSampleSize = ut.effective_sample_size(data[weightingAttribute])
        # p_subgroup = positivesSubgroup / instancesSubgroup
        # p_dataset = positivesDataset / instancesDataset

        negativesSubgroup = instancesSubgroup - positivesSubgroup
        negativesDataset = instancesDataset - positivesDataset
        positivesComplement = positivesDataset - positivesSubgroup
        negativesComplement = negativesDataset - negativesSubgroup
        val = scipy.stats.chi2_contingency([[positivesSubgroup, positivesComplement],
                                            [negativesSubgroup, negativesComplement]], correction=True) [0]
        return scipy.stats.chi2.sf(val * effectiveSampleSize / instancesDataset, 1)
    
    def __init__(self, direction='bidirect', min_instances=5):
        if (direction == 'bidirect'):
            self.bidirect = True
            self.direction_positive = True
        if (direction == 'positive'):
            self.bidirect = False
            self.direction_positive = True
        if (direction == 'negative'):
            self.bidirect = False
            self.direction_positive = False
        self.min_instances = min_instances

    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):

        if not self.isApplicable(subgroup):
            raise BaseException("Quality measure cannot be used for this target class")
        if weightingAttribute is None:
            result = self.evaluateFromStatistics (*subgroup.get_base_statistics(data))
        else:
            (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup) = subgroup.get_base_statistics(data, weightingAttribute)
            weights = data[weightingAttribute]
            base = self.evaluateFromStatistics (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup)           
            result = base * ut.effective_sample_size(weights) / instancesDataset
        return result
    
    def evaluateFromStatistics (self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return ChiSquaredQF.chiSquaredQF (instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup, self.min_instances, self.bidirect, self.direction_positive)

    def supportsWeights(self):
        return True
    
    def isApplicable(self, subgroup):
        return isinstance(subgroup.target, NominalTarget)
    
class StandardQF (AbstractInterestingnessMeasure, BoundedInterestingnessMeasure):
    @staticmethod     
    def standardQF (a, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        if (instancesSubgroup == 0):
            return 0
        p_subgroup = positivesSubgroup / instancesSubgroup
        p_dataset = positivesDataset / instancesDataset
        return (instancesSubgroup / instancesDataset) ** a * (p_subgroup - p_dataset)
            
    def __init__(self, a):
        self.a = a
    
    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):
        if not self.isApplicable(subgroup):
            raise BaseException("Quality measure cannot be used for this target class")
        return self.evaluateFromStatistics (*subgroup.get_base_statistics(data, weightingAttribute))
    
    def optimisticEstimateFromDataset(self, data, subgroup, weightingAttribute=None):
        if not self.isApplicable(subgroup):
            raise BaseException("Quality measure cannot be used for this target class")
        return self.optimisticEstimateFromStatistics (*subgroup.get_base_statistics(data, weightingAttribute))

    def evaluateFromStatistics (self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return StandardQF.standardQF (self.a, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup)
    
    def optimisticEstimateFromStatistics (self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return StandardQF.standardQF (self.a, instancesDataset, positivesDataset, positivesSubgroup, positivesSubgroup)

    def supportsWeights(self):
        return True
    
    def isApplicable(self, subgroup):
        return isinstance(subgroup.target, NominalTarget)

class WRAccQF (StandardQF):
    def __init__(self, a):
        super().__init__(a)
        self.a = 1.0
        
class LiftQF (StandardQF):
    def __init__(self, a):
        super().__init__(a)
        self.a = 0.0
        
class SimpleBinomial(StandardQF):
    def __init__(self, a):
        super().__init__(a)
        self.a = 0.5
        
#####
# GeneralizationAware Interestingness Measures
#####
class GAStandardQF (AbstractInterestingnessMeasure):    
    def __init__(self, a):
        self.a = a

    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):
        (instancesDataset, _, instancesSubgroup, positivesSubgroup) = subgroup.get_base_statistics(data, weightingAttribute)
        if ((instancesSubgroup == 0) or (instancesDataset == instancesSubgroup)):
            return 0
        p_subgroup = positivesSubgroup / instancesSubgroup
        maxTargetShare = getMaxGeneralizationTargetShare(data, subgroup, weightingAttribute)
        relativeSize = (instancesSubgroup / instancesDataset)
        return relativeSize ** self.a * (p_subgroup - maxTargetShare)

    def supportsWeights(self):
        return True

    def isApplicable(self, subgroup):
        return isinstance(subgroup.target, NominalTarget)


def getMaxGeneralizationTargetShare(data, subgroup, weightingAttribute=None):
    selectors = subgroup.subgroupDescription.selectors
    generalizations = ut.powerset(selectors)
    maxTargetShare = 0
    for sels in generalizations:
        sgd = SubgroupDescription(list(sels))
        sg = Subgroup(subgroup.target, sgd)
        (_, _, instancesSubgroup, positivesSubgroup) = sg.get_base_statistics(data, weightingAttribute)
        targetShare = positivesSubgroup / instancesSubgroup
        maxTargetShare = max(maxTargetShare, targetShare)
    return maxTargetShare







'''
Created on 02.05.2016

@author: lemmerfn
'''
from functools import partial
from heapq import heappush, heappop
import itertools
import numpy as np
import pandas as pd

all_statistics = ('size_sg', 'size_dataset', 'positives_sg', 'positives_dataset', 'size_complement', 'relative_size_sg', 'relative_size_complement', 'coverage_sg', 'coverage_complement', 'target_share_sg', 'target_share_complement', 'target_share_dataset', 'lift')
all_statistics_weighted = all_statistics + ('size_sg_weighted', 'size_dataset_weighted', 'positives_sg_weighted', 'positives_dataset_weighted', 'size_complement_weighted', 'relative_size_sg_weighted', 'relative_size_complement_weighted', 'coverage_sg_weighted', 'coverage_complement_weighted', 'target_share_sg_weighted', 'target_share_complement_weighted', 'target_share_dataset_weighted', 'lift_weighted')
all_statistics_numeric = ('size_sg', 'size_dataset', 'mean_sg', 'mean_dataset', 'std_sg', 'std_dataset', 'median_sg', 'median_dataset', 'max_sg', 'max_dataset', 'min_sg', 'min_dataset', 'mean_lift', 'median_lift')


def addIfRequired (result, sg, quality, task, check_for_duplicates=False):
    if (quality > task.minQuality):
        if (check_for_duplicates and (quality, sg) in result):
            return
        if (len(result) < task.resultSetSize):
            heappush(result, (quality, sg))
        elif (quality > result[0][0]):
            
            heappop(result)
            heappush(result, (quality, sg))


def minimumRequiredQuality (result, task):
    if (len(result) < task.resultSetSize):
        return task.minQuality
    else:
        return result [0][0]
    
# Returns the cutpoints for discretization
def equalFrequencyDiscretization (data, attributeName, nbins=5, weightingAttribute=None):
    cutpoints = []
    if weightingAttribute == None:
        cleanedData = data[attributeName]
        cleanedData = cleanedData[~np.isnan(cleanedData)]
        sortedData = sorted(cleanedData)
        numberInstances = len(sortedData)
        for i in range(1, nbins):
            position = i * numberInstances // nbins
            while True:
                if (position >= numberInstances):
                    break
                val = sortedData [position]
                if (not val in cutpoints):
                    break
                position += 1
            # print (sortedData [position])
            if (not val in cutpoints):
                cutpoints.append(val)
    else:
        cleanedData = data[[attributeName, weightingAttribute]]
        cleanedData = cleanedData[~np.isnan(cleanedData[attributeName])]
        cleanedData.sort(order=attributeName)
    
        overall_weights = cleanedData[weightingAttribute].sum()
        remaining_weights = overall_weights
        bin_size = overall_weights / nbins
        sum_of_weights = 0
        for row in cleanedData:
            sum_of_weights += row[weightingAttribute]
            if (sum_of_weights > bin_size):
                if (not row[attributeName] in cutpoints):
                    cutpoints.append(row[attributeName])
                    remaining_weights = remaining_weights - sum_of_weights
                    if remaining_weights < 1.5 * (bin_size):
                        break
                    sum_of_weights = 0
    return cutpoints

def printResultSet(data, result, statisticsToShow, weightingAttribute=None, printHeader=True, includeTarget=False):
    if printHeader:
        s = "Quality\tSubgroup"
        for stat in statisticsToShow:
            s += "\t" + stat
        print (s) 
    for (q, sg) in result:
        sg.calculateStatistics(data, weightingAttribute)
        s = str(q) + ":\t" + str(sg.subgroupDescription) 
        if includeTarget:
            s += str(sg.target)
        for stat in statisticsToShow:
            s += "\t" + str(sg.statistics[stat])
        print (s) 

def resultAsTable(data, result, statisticsToShow, weightingAttribute=None, printHeader=True, includeTarget=False):
    table = []
    if printHeader:
        row = ["quality", "subgroup"]
        for stat in statisticsToShow:
            row.append(stat)
        table.append(row)
    for (q, sg) in result:
        sg.calculateStatistics(data, weightingAttribute)
        row = [str(q), str(sg.subgroupDescription)]
        if includeTarget:
            row.append(str(sg.target))
        for stat in statisticsToShow:
            row.append(str(sg.statistics[stat]))
        table.append(row) 
    return table

def resultsAsDataFrame (data, result, statisticsToShow=all_statistics, autoround=False, weightingAttribute=None, includeTarget=False):
    res = resultAsTable(data, result, statisticsToShow, weightingAttribute, True, includeTarget)
    headers = res.pop(0)
    df = pd.DataFrame(res, columns=headers, dtype=np.float64)
    if (autoround):
        df = results_df_autoround(df)
    return df

as_df = resultsAsDataFrame

def conditional_invert (val, invert):
    return - 2* (invert -0.5) * val


def results_df_autoround (df):
    return df.round({
                'quality' : 3,
                'size_sg' : 0,
                'size_dataset' : 0,
                'positives_sg': 0,
                'positives_dataset':0,
                'size_complement' : 0,
                'relative_size_sg':3,
                'relative_size_complement':3,
                'coverage_sg':3,
                'coverage_complement':3,
                'target_share_sg':3,
                'target_share_complement':3,
                'target_share_dataset':3,
                'lift':3,
                  
                'size_sg_weighted': 1,
                'size_dataset_weighted':1,
                'positives_sg_weighted':1,
                'positives_dataset_weighted':1,
                'size_complement_weighted':1,
                'relative_size_sg_weighted':3,
                'relative_size_complement_weighted':3,
                'coverage_sg_weighted':3,
                'coverage_complement_weighted':3,
                'target_share_sg_weighted':3,
                'target_share_complement_weighted':3,
                'target_share_dataset_weighted':3,
                'lift_weighted':3})


def perc_formatter (x):
    return "{0:.1f}%".format(x * 100)

def float_formatter (x, digits=2):
    return ("{0:." + str(digits) + "f}").format(x)

def to_latex (data, result, statistics_to_show):
    df = resultsAsDataFrame (data, result) [statistics_to_show]
    latex = df.to_latex(index=False, col_space=10, formatters={
                'quality' : partial (float_formatter, digits=3),
                'size_sg' : partial (float_formatter, digits=0),
                'size_dataset' : partial (float_formatter, digits=0),
                'positives_sg': partial (float_formatter, digits=0),
                'positives_dataset':partial (float_formatter, digits=0),
                'size_complement' : partial (float_formatter, digits=0),
                'relative_size_sg':perc_formatter,
                'relative_size_complement':perc_formatter,
                'coverage_sg':perc_formatter,
                'coverage_complement':perc_formatter,
                'target_share_sg':perc_formatter,
                'target_share_complement':perc_formatter,
                'target_share_dataset':perc_formatter,
                'lift':partial (float_formatter, digits=1),
                  
                'size_sg_weighted': partial (float_formatter, digits=1),
                'size_dataset_weighted':partial (float_formatter, digits=1),
                'positives_sg_weighted':partial (float_formatter, digits=1),
                'positives_dataset_weighted':partial (float_formatter, digits=1),
                'size_complement_weighted':partial (float_formatter, digits=1),
                'relative_size_sg_weighted':perc_formatter,
                'relative_size_complement_weighted':perc_formatter,
                'coverage_sg_weighted': perc_formatter,
                'coverage_complement_weighted' : perc_formatter,
                'target_share_sg_weighted': perc_formatter,
                'target_share_complement_weighted': perc_formatter,
                'target_share_dataset_weighted': perc_formatter,
                'lift_weighted': perc_formatter}
    )
    latex = latex.replace (' AND ', ' $\wedge$ ')
    return latex


def isCategoricalAttribute (data, attribute_name):    
    return attribute_name in data.select_dtypes(exclude=['number']).columns.values

def isNumericalAttribute (data, attribute_name):
    return attribute_name in data.select_dtypes(include=['number']).columns.values

def remove_selectors_with_attributes(selector_list, attribute_list):
    return [x for x in selector_list if not x.attributeName in attribute_list]

def effective_sample_size(weights):
    return sum(weights) ** 2 / sum(weights ** 2)

# from https://docs.python.org/3/library/itertools.html#recipes
def powerset(iterable):
        "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
        s = list(iterable)
        return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s)))

def overlap (sg, anotherSG, data):
    cover_sg = sg.covers(data)
    cover_another_sg = anotherSG.covers(data)
    union = np.logical_or(cover_sg, cover_another_sg)
    intercept = np.logical_and(cover_sg, cover_another_sg)
    sim = np.sum(intercept) / np.sum(union)
    return sim

#####
# bitset operations
#####
def to_bits (list_of_ints):
    v = 0
    for x in list_of_ints:
        v += 1 << x
    return v


def count_bits (bitset_as_int):
    c = 0
    while bitset_as_int > 0:
        c += 1
        bitset_as_int &= bitset_as_int - 1
    return c


def find_set_bits(bitset_as_int):
    while bitset_as_int > 0:
        x = bitset_as_int.bit_length() - 1
        yield x
        bitset_as_int = bitset_as_int - (1 << x)


#####
# TID-list operations
#####
def intersect_of_ordered_list(list_1, list_2):
    result = []
    i = 0
    j = 0
    while i < len(list_1) and j < len(list_2):
        if list_1[i] < list_2[j]:
                i += 1
        elif list_2[j] < list_1[i]:
            j+= 1
        else:
            result.append(list_1[i])
            j += 1
            i += 1
    return result


import pysubgroup as ps
import pandas as pd
import numpy as np

import pprint
pp = pprint.PrettyPrinter(indent=4)

data = np.array ([[1, 2, 3, 4, 5], ["F", "F", "F", "Tr", "Tr"]]).T
data = pd.DataFrame(data, columns=["Target", "A"])
data ["Target"] = pd.to_numeric (data["Target"])


target = ps.NumericTarget ('Target')
print (data[target.target_variable])
sg = ps.Subgroup (target, ps.NominalSelector("A", "Tr"))
print (target.get_base_statistics(data, sg))
sg.calculateStatistics(data)
# pp.pprint (sg.statistics)
qf = ps.StandardQF_numeric (1.0)
print (qf.optimisticEstimateFromDataset(data, sg))



'''
Created on 29.09.2017

@author: lemmerfn
'''
from functools import total_ordering
from pysubgroup.measures import AbstractInterestingnessMeasure, \
    BoundedInterestingnessMeasure

@total_ordering
class FITarget(object):
    def __init__(self):
        pass
        
    def __repr__(self):
        return "T: Frequent Itemsets"
    
    def __eq__(self, other): 
        return self.__dict__ == other.__dict__
    
    def __lt__(self, other): 
        return str(self) < str(other)
    
    def getAttributes(self):
        return []
    
    def get_base_statistics (self, data, subgroup, weightingAttribute=None): 
        if (weightingAttribute == None):
            sg_instances = subgroup.subgroupDescription.covers(data)
            return sg_instances.sum()  
        else:
            raise NotImplemented("Attribute weights with numeric targets are not yet implemented.")
        
    def calculateStatistics (self, subgroup, data, weightingAttribute=None):
        if weightingAttribute != None:
            raise NotImplemented("Attribute weights with numeric targets are not yet implemented.")
        sg_instances = subgroup.subgroupDescription.covers(data)
        
        subgroup.statistics['size_sg'] = len(sg_instances)
        subgroup.statistics['size_dataset'] = len (data)
        
class CountQF (AbstractInterestingnessMeasure, BoundedInterestingnessMeasure):
    def __init__(self):
        pass
    
    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):
        return subgroup.subgroupDescription.covers(data).sum()
    
    def optimisticEstimateFromDataset(self, data, subgroup):
        return subgroup.subgroupDescription.covers(data).sum()
        
    def evaluateFromStatistics (self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return instancesSubgroup
    
    def optimisticEstimateFromStatistics (self, instancesDataset, positivesDataset, instancesSubgroup, positivesSubgroup):
        return instancesSubgroup

    def isApplicable(self, subgroup):
        return isinstance(subgroup.target, FITarget)

    def supportsWeights(self):
        return False


class AreaQF(AbstractInterestingnessMeasure):
    def __init__(self):
        pass

    def evaluateFromDataset(self, data, subgroup, weightingAttribute=None):
        return len(subgroup.subgroupDescription) * subgroup.subgroupDescription.covers(data).sum()

    def isApplicable(self, subgroup):
        return isinstance(subgroup.target, FITarget)

    def supportsWeights(self):
        return False



import pysubgroup as ps
import pandas as pd

import pprint

pp = pprint.PrettyPrinter(indent=4)

data = pd.read_csv("~/datasets/titanic.csv")
searchSpace = ps.createSelectors(data, ignore="survived")
dt = data.dtypes

task = ps.SubgroupDiscoveryTask (data, ps.FITarget, searchSpace, resultSetSize=10, depth=5, qf=ps.CountQF())
result = ps.SimpleDFS().execute(task)

for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))


task = ps.SubgroupDiscoveryTask (data, ps.FITarget, searchSpace, resultSetSize=10, depth=3, qf=ps.AreaQF())
result = ps.SimpleDFS().execute(task)

for (q, sg) in result:
    print (f"{q}\t{sg.subgroupDescription}\t{sg.count(data)}")


import pysubgroup as ps
import pandas as pd
from scipy.io import arff
from timeit import default_timer as timer


data = pd.DataFrame (arff.loadarff("../data/credit-g.arff") [0])
target = ps.NumericTarget ('credit_amount')
searchSpace = ps.createNominalSelectors(data, ignore=['credit_amount'])

task = ps.SubgroupDiscoveryTask (data, target, searchSpace, resultSetSize=10, depth=3, qf=ps.StandardQF_numeric(1, False))
print (searchSpace)

start = timer()
result = ps.SimpleDFS().execute(task)
end = timer()
print(f"Time elapsed: {end - start}")
for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))



from scipy.io import arff
import pysubgroup as ps
import pandas as pd

import cProfile
from timeit import default_timer as timer

data = pd.DataFrame (arff.loadarff("../data/credit-g.arff") [0])

target = ps.NominalTarget ('class', b'bad')
searchSpace = ps.createNominalSelectors(data, ignore=['class'])
task = ps.SubgroupDiscoveryTask (data, target, searchSpace, resultSetSize=10, depth=5, qf=ps.StandardQF(1.0))

start = timer()
result = ps.Apriori().execute(task)
end = timer()
print("Time elapsed: ", (end - start)) 


for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))   

print ("******")

start = timer()
result = ps.SimpleDFS().execute(task)
end = timer()
print("Time elapsed: ", (end - start)) 
for (q, sg) in result:
    print (str(q) + ":\t" + str(sg.subgroupDescription))   



from setuptools import setup

setup(
    name='pysubgroup',
    version='0.5.3',
    packages=['', 'test'],
    package_dir={'': 'pysubgroup', 'test': "pysubgroup.tests"},
    url='http://florian.lemmerich.net/pysubgroup',
    license='',
    author='Florian Lemmerich',
    author_email='florian@lemmerich.net',
    description='pysubgroup is a Python library for the data analysis task of subgroup discovery.',
    install_requires=[
              'pandas','scipy','numpy','matplotlib'
          ],
    python_requires='>=3.5'
)



from functools import partial

import pandas as pd
import numpy as np
from scipy.cluster.hierarchy import dendrogram, linkage
from scipy.spatial.distance import squareform
from matplotlib import pyplot as plt

import pysubgroup.boolean_target as bt
import pysubgroup.utils as ut

def plot_sgbars (sgs, shares_sg, shares_compl, sg_relative_sizes, ylabel="target share", title="Discovered Subgroups", dynamic_widths=False, suffix=""):
    x = np.arange (len(sgs))

    base_width = 0.8
    if dynamic_widths:
        width_sg = 0.02 + base_width * sg_relative_sizes
        width_compl = base_width - width_sg
    else:
        width_sg = base_width / 2
        width_compl = base_width / 2
    
    fig, ax = plt.subplots()
    rects1 = ax.bar(x, shares_sg, width_sg, align='edge')
    rects2 = ax.bar(x + width_sg, shares_compl, width_compl, align='edge', color='#61b76f')

    ax.set_ylabel(ylabel)
    ax.set_title(title)
    ax.set_xticks(x + base_width / 2)
    ax.set_xticklabels(sgs, rotation=90)
    

    ax.legend((rects1[0], rects2[0]), ('subgroup', 'complement'))
    fig.set_size_inches(12, len(sgs))
    
    return fig


def plot_roc(result_df, data, qf=bt.StandardQF(0.5), levels=40, annotate=False):
    instances_dataset = len(data)
    positives_dataset = np.max(result_df['positives_dataset'])
    negatives_dataset = instances_dataset - positives_dataset

    xlist = np.linspace(0.01, 0.99, 100)
    ylist = np.linspace(0.01, 0.99, 100)
    X, Y = np.meshgrid(xlist, ylist)
    f = np.vectorize(partial(qf.evaluateFromStatistics, instances_dataset, positives_dataset), otypes=[np.float])
    Z = f(X * negatives_dataset + Y * positives_dataset, Y * positives_dataset)
    max_val = np.max([np.max(Z), -np.min(Z)])

    fig, ax = plt.subplots()
    cm = plt.cm.bwr

    plt.contourf(X, Y, Z, levels, cmap=cm, vmin=-max_val, vmax=max_val)

    for i, sg in result_df.iterrows():
        rel_positives_sg = sg['positives_sg'] / positives_dataset
        rel_negatives_sg = (sg['size_sg'] - sg['positives_sg']) / negatives_dataset
        ax.plot(rel_negatives_sg, rel_positives_sg, 'ro', color='black')
        if annotate:
            label_margin = 0.01
            ax.annotate(str(i), (rel_negatives_sg + label_margin, rel_positives_sg + label_margin))

    # plt.colorbar(cp)
    plt.title('Discovered subgroups')
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')

    return fig

def plot_npspace(result_df, data, annotate=True, fixed_limits=False):

    fig, ax = plt.subplots()

    for i, sg in result_df.iterrows():
        target_share_sg = sg['target_share_sg']
        size_sg = sg['size_sg']
        ax.plot(size_sg, target_share_sg, 'ro', color='black')
        if annotate:
            ax.annotate(str(i), (size_sg+5, target_share_sg+0.001))

    if fixed_limits:
        plt.xlim((0,len(data)))
        plt.ylim((0, 1))

    plt.title('Discovered subgroups')
    plt.xlabel('Size of Subgroup')
    plt.ylabel('Target Share Subgroup')

    return fig


def plot_distribution_numeric(sg, data, bins):
    fig, ax = plt.subplots()
    target_values_sg = data[sg.covers(data)][sg.target.getAttributes()].values
    target_values_data = data[sg.target.getAttributes()].values
    plt.hist(target_values_sg, bins, alpha=0.5, label=str(sg.subgroupDescription), density=True)
    plt.hist(target_values_data, bins, alpha=0.5, label="Overall Data", density=True)
    plt.legend(loc='upper right')
    return fig


def compare_distributions_numeric(sgs, data, bins):
    fig, ax = plt.subplots()
    for sg in sgs:
        target_values_sg = data[sg.covers(data)][sg.target.getAttributes()].values
        plt.hist(target_values_sg, bins, alpha=0.3, label=str(sg.subgroupDescription), density=True)
    plt.legend(loc='upper right')
    return fig

def similarity_sgs (sgd_results, data, color = True):
    sgs = [x[1] for x in sgd_results]
    #sgNames = [str(sg.subgroupDescription) for sg in sgs]
    dists = [[ut.overlap(sg, sg2, data) for sg2 in sgs] for sg in sgs]
    dist_df = pd.DataFrame(dists)
    if color:
        dist_df = dist_df.style.background_gradient()
    return dist_df

def similarity_dendrogram(result, data):
    fig, ax = plt.subplots()
    dist_df = similarity_sgs(result, data, color=False)
    mat = 1 - dist_df.values
    dists = squareform(mat)
    linkage_matrix = linkage(dists, "single")
    dendrogram(linkage_matrix, labels=dist_df.index)
    return fig





import pysubgroup as ps
import pandas as pd
import matplotlib.pyplot as plt
plt.interactive(False)


data = pd.read_csv("~/datasets/titanic.csv")
target = ps.NominalTarget ('survived', 0)
searchSpace = ps.createSelectors(data, ignore=['survived'])
task = ps.SubgroupDiscoveryTask (data, target, searchSpace, 
                                 resultSetSize=5, depth=2, 
                                 qf=ps.ChiSquaredQF())

result = ps.SimpleDFS().execute(task)

dfs = ps.utils.resultsAsDataFrame (data, result)
plt = ps.plot_roc (data, dfs, ps.ChiSquaredQF())
plt.show()